/*

Copyright (c) 2001-2007 Michael Terry
Copyright (c) 2009 Paul Ivanov
Copyright (c) 2011 Sergei Riaguzov
Copyright (c) 2011 Dennis Hilmar
Copyright (c) 2011 OBATA Akio
Copyright (c) 2013-2014 Arthur Borsboom

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "../config.h"
#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include "fio.h"
#include "help.h"
#include "xpad-app.h"
#include "xpad-pad.h"
#include "xpad-pad-properties.h"
#include "xpad-periodic.h"
#include "xpad-preferences.h"
#include "xpad-text-buffer.h"
#include "xpad-text-view.h"
#include "xpad-toolbar.h"
#include "xpad-tray.h"

struct XpadPadPrivate
{
	/* saved values */
	gint x, y;
	guint width, height;
	gboolean location_valid;
	gchar *infoname;
	gchar *contentname;
	gboolean sticky;
	
	/* selected child widgets */
	GtkWidget *textview;
	GtkWidget *scrollbar;
	
	/* toolbar stuff */
	GtkWidget *toolbar;
	guint toolbar_timeout;
	guint toolbar_height;
	gboolean toolbar_expanded;
	gboolean toolbar_pad_resized;
	
	/* properties window */
	GtkWidget *properties;
	
	/* menus */
	GtkWidget *menu;
	GtkWidget *highlight_menu;

	gboolean unsaved_content;
	gboolean unsaved_info;

	GtkClipboard *clipboard;
	
	XpadPadGroup *group;
};

G_DEFINE_TYPE_WITH_PRIVATE(XpadPad, xpad_pad, GTK_TYPE_WINDOW)

enum
{
	CLOSED,
	LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_GROUP,
  LAST_PROP
};

static void xpad_pad_load_info (XpadPad *pad, gboolean *show);
static GtkWidget *menu_get_popup_highlight (XpadPad *pad, GtkAccelGroup *accel_group);
static GtkWidget *menu_get_popup_no_highlight (XpadPad *pad, GtkAccelGroup *accel_group);
static void xpad_pad_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void xpad_pad_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void xpad_pad_dispose (GObject *object);
static void xpad_pad_finalize (GObject *object);
static void xpad_pad_show (XpadPad *pad);
static gboolean xpad_pad_configure_event (XpadPad *pad, GdkEventConfigure *event);
static gboolean xpad_pad_toolbar_size_allocate (XpadPad *pad, GtkAllocation *event);
static gboolean xpad_pad_delete_event (XpadPad *pad, GdkEvent *event);
static gboolean xpad_pad_popup_menu (XpadPad *pad);
static void xpad_pad_popup_deactivate (GtkWidget *menu, XpadPad *pad);
static gboolean xpad_pad_button_press_event (XpadPad *pad, GdkEventButton *event);
static gboolean xpad_pad_text_view_button_press_event (GtkWidget *text_view, GdkEventButton *event, XpadPad *pad);
static void xpad_pad_text_changed (XpadPad *pad, GtkTextBuffer *buffer);
static void xpad_pad_notify_has_scrollbar (XpadPad *pad);
static void xpad_pad_notify_has_decorations (XpadPad *pad);
static void xpad_pad_notify_has_toolbar (XpadPad *pad);
static void xpad_pad_notify_autohide_toolbar (XpadPad *pad);
static void xpad_pad_hide_toolbar (XpadPad *pad);
static void xpad_pad_show_toolbar (XpadPad *pad);
static void xpad_pad_popup (XpadPad *pad, GdkEventButton *event);
static void xpad_pad_spawn (XpadPad *pad);
static void xpad_pad_clear (XpadPad *pad);
static void xpad_pad_undo (XpadPad *pad);
static void xpad_pad_redo (XpadPad *pad);
static void xpad_pad_cut (XpadPad *pad);
static void xpad_pad_copy (XpadPad *pad);
static void xpad_pad_paste (XpadPad *pad);
static void xpad_pad_delete (XpadPad *pad);
static void xpad_pad_open_properties (XpadPad *pad);
static void xpad_pad_open_preferences ();
static void xpad_pad_quit ();
/* static void xpad_pad_show_all (XpadPad *pad); */
static void xpad_pad_close_all (XpadPad *pad);
static void xpad_pad_sync_title (XpadPad *pad);
static void xpad_pad_set_group (XpadPad *pad, XpadPadGroup *group);
static gboolean xpad_pad_leave_notify_event (GtkWidget *pad, GdkEventCrossing *event);
static gboolean xpad_pad_enter_notify_event (GtkWidget *pad, GdkEventCrossing *event);
static void xpad_pad_toolbar_popup (GtkWidget *toolbar, GtkMenu *menu, XpadPad *pad);
static void xpad_pad_toolbar_popdown (GtkWidget *toolbar, GtkMenu *menu, XpadPad *pad);
static XpadPadGroup *xpad_pad_get_group (XpadPad *pad);

static guint signals[LAST_SIGNAL] = { 0 };

/* Create a new empty pad. */
GtkWidget *
xpad_pad_new (XpadPadGroup *group)
{
	return GTK_WIDGET (g_object_new (XPAD_TYPE_PAD, "group", group, NULL));
}

/* Create a new pad based on the provided info-xxxxx file from the config directory and return this pad */
GtkWidget *
xpad_pad_new_with_info (XpadPadGroup *group, const gchar *info_filename, gboolean *show)
{
	GtkWidget *pad = GTK_WIDGET (g_object_new (XPAD_TYPE_PAD, "group", group, NULL));
	
	XPAD_PAD (pad)->priv->infoname = g_strdup (info_filename);
	xpad_pad_load_info (XPAD_PAD (pad), show);
	xpad_pad_load_content (XPAD_PAD (pad));
	gtk_window_set_role (GTK_WINDOW (pad), XPAD_PAD (pad)->priv->infoname);
	
	return pad;
}

/* Create a new pad based on the provided filename from the command line */
GtkWidget *
xpad_pad_new_from_file (XpadPadGroup *group, const gchar *filename)
{
	GtkWidget *pad = NULL;
	gchar *content;
	
	content = fio_get_file (filename);
	
	if (!content)
	{
		gchar *usertext = g_strdup_printf (_("Could not read file %s."), filename);
		xpad_app_error (NULL, usertext, NULL);
		g_free (usertext);
	}
	else
	{
		GtkTextBuffer *buffer;
		
		pad = GTK_WIDGET (g_object_new (XPAD_TYPE_PAD, "group", group, NULL));
		buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (XPAD_PAD (pad)->priv->textview));

		xpad_text_buffer_freeze_undo (XPAD_TEXT_BUFFER (buffer));
		g_signal_handlers_block_by_func (buffer, xpad_pad_text_changed, pad);
		
		xpad_text_buffer_set_text_with_tags (XPAD_TEXT_BUFFER (buffer), content ? content : "");
		g_free (content);
		
		g_signal_handlers_unblock_by_func (buffer, xpad_pad_text_changed, pad);
		xpad_text_buffer_thaw_undo (XPAD_TEXT_BUFFER (buffer));

		xpad_pad_text_changed(XPAD_PAD(pad), buffer);
	}
	
	return pad;
}

/* Class pad - constructor */
static void
xpad_pad_class_init (XpadPadClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	
	gobject_class->dispose = xpad_pad_dispose;
	gobject_class->finalize = xpad_pad_finalize;
	gobject_class->set_property = xpad_pad_set_property;
	gobject_class->get_property = xpad_pad_get_property;
	
	signals[CLOSED] =
		g_signal_new ("closed",
						  G_OBJECT_CLASS_TYPE (gobject_class),
						  G_SIGNAL_RUN_FIRST,
						  G_STRUCT_OFFSET (XpadPadClass, closed),
						  NULL, NULL,
						  g_cclosure_marshal_VOID__VOID,
						  G_TYPE_NONE,
						  0);
	
	/* Properties */
	
	g_object_class_install_property (gobject_class,
												PROP_GROUP,
												g_param_spec_pointer ("group",
																			 "Pad Group",
																			 "Pad group for this pad",
																			 G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
}

/* Class pad - initializer */
static void
xpad_pad_init (XpadPad *pad)
{
	GtkWidget *vbox;
	GtkAccelGroup *accel_group;

	pad->priv = xpad_pad_get_instance_private(pad);

	pad->priv->x = 0;
	pad->priv->y = 0;
	pad->priv->location_valid = FALSE;
	pad->priv->infoname = NULL;
	pad->priv->contentname = NULL;
	pad->priv->textview = NULL;
	pad->priv->scrollbar = NULL;
	pad->priv->toolbar = NULL;
	pad->priv->toolbar_timeout = 0;
	pad->priv->toolbar_height = 0;
	pad->priv->toolbar_expanded = FALSE;
	pad->priv->toolbar_pad_resized = TRUE;
	pad->priv->properties = NULL;
	pad->priv->unsaved_content = FALSE;
	pad->priv->unsaved_info = FALSE;
	pad->priv->group = NULL;
	g_object_get (xpad_global_settings,
			"width", &pad->priv->width,
			"height", &pad->priv->height,
			"autostart-sticky", &pad->priv->sticky, NULL);

	XpadTextView *text_view = XPAD_TEXT_VIEW (xpad_text_view_new ());
	xpad_text_view_set_pad (text_view, pad);

	pad->priv->textview = GTK_WIDGET (text_view);

	pad->priv->scrollbar = GTK_WIDGET (g_object_new (GTK_TYPE_SCROLLED_WINDOW,
		"hadjustment", NULL,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_NONE,
		"vadjustment", NULL,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"child", pad->priv->textview,
		NULL));

	pad->priv->toolbar = GTK_WIDGET ( xpad_toolbar_new (pad));

	accel_group = gtk_accel_group_new ();
	gtk_window_add_accel_group (GTK_WINDOW (pad), accel_group);

	pad->priv->menu = menu_get_popup_no_highlight (pad, accel_group);
	pad->priv->highlight_menu = menu_get_popup_highlight (pad, accel_group);

	gtk_accel_group_connect (accel_group, GDK_KEY_Q, GDK_CONTROL_MASK, 0,
									 g_cclosure_new_swap (G_CALLBACK (xpad_app_quit), pad, NULL));
	g_object_unref (G_OBJECT (accel_group));

	vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
	gtk_box_set_homogeneous (GTK_BOX(vbox), FALSE);
	gtk_box_pack_start (GTK_BOX(vbox), pad->priv->scrollbar, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX(vbox), pad->priv->toolbar, FALSE, FALSE, 0);

	gtk_container_child_set (GTK_CONTAINER (vbox), pad->priv->toolbar, "expand", FALSE, NULL);

	gboolean decorations;
	g_object_get (xpad_global_settings, "has-decorations", &decorations, NULL);
	gtk_window_set_decorated (GTK_WINDOW(pad), decorations);
	gtk_window_set_default_size (GTK_WINDOW(pad), (gint) pad->priv->width, (gint) pad->priv->height);
	gtk_window_set_gravity (GTK_WINDOW(pad),  GDK_GRAVITY_STATIC); /* static gravity makes saving pad x,y work */
	gtk_window_set_skip_pager_hint (GTK_WINDOW(pad), decorations);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(pad), !decorations);
	gtk_window_set_position (GTK_WINDOW(pad), GTK_WIN_POS_MOUSE);

	g_object_set (G_OBJECT (pad),
		"child", vbox,
		NULL);

	xpad_pad_notify_has_scrollbar (pad);
	xpad_pad_notify_has_selection (pad);
	xpad_pad_notify_clipboard_owner_changed (pad);
	xpad_pad_notify_undo_redo_changed (pad);

 	pad->priv->clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);

	if (pad->priv->sticky)
		gtk_window_stick (GTK_WINDOW (pad));
	else
		gtk_window_unstick (GTK_WINDOW (pad));

	xpad_pad_sync_title (pad);

	gtk_widget_show_all (vbox);

	gtk_widget_hide (pad->priv->toolbar);
	xpad_pad_notify_has_toolbar (pad);

	/* Set up signals */
	gtk_widget_add_events (GTK_WIDGET (pad), GDK_BUTTON_PRESS_MASK | GDK_PROPERTY_CHANGE_MASK);
	gtk_widget_add_events (pad->priv->toolbar, GDK_ALL_EVENTS_MASK);
	g_signal_connect (pad->priv->textview, "button-press-event", G_CALLBACK (xpad_pad_text_view_button_press_event), pad);
	g_signal_connect_swapped (pad->priv->textview, "popup-menu", G_CALLBACK (xpad_pad_popup_menu), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "size-allocate", G_CALLBACK (xpad_pad_toolbar_size_allocate), pad);
	g_signal_connect (pad, "button-press-event", G_CALLBACK (xpad_pad_button_press_event), NULL);
	g_signal_connect (pad, "configure-event", G_CALLBACK (xpad_pad_configure_event), NULL);
	g_signal_connect (pad, "delete-event", G_CALLBACK (xpad_pad_delete_event), NULL);
	g_signal_connect (pad, "popup-menu", G_CALLBACK (xpad_pad_popup_menu), NULL);
	g_signal_connect (pad, "show", G_CALLBACK (xpad_pad_show), NULL);
	g_signal_connect_swapped (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)), "changed", G_CALLBACK (xpad_pad_text_changed), pad);

	g_signal_connect (pad, "enter-notify-event", G_CALLBACK (xpad_pad_enter_notify_event), NULL);
	g_signal_connect (pad, "leave-notify-event", G_CALLBACK (xpad_pad_leave_notify_event), NULL);

	g_signal_connect_swapped (xpad_global_settings, "notify::has-decorations", G_CALLBACK (xpad_pad_notify_has_decorations), pad);
	g_signal_connect_swapped (xpad_global_settings, "notify::has-toolbar", G_CALLBACK (xpad_pad_notify_has_toolbar), pad);
	g_signal_connect_swapped (xpad_global_settings, "notify::autohide-toolbar", G_CALLBACK (xpad_pad_notify_autohide_toolbar), pad);
	g_signal_connect_swapped (xpad_global_settings, "notify::has-scrollbar", G_CALLBACK (xpad_pad_notify_has_scrollbar), pad);
	g_signal_connect_swapped (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)), "notify::has-selection", G_CALLBACK (xpad_pad_notify_has_selection), pad);
	g_signal_connect_swapped (pad->priv->clipboard, "owner-change", G_CALLBACK (xpad_pad_notify_clipboard_owner_changed), pad);

	g_signal_connect_swapped (pad->priv->toolbar, "activate-new", G_CALLBACK (xpad_pad_spawn), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-clear", G_CALLBACK (xpad_pad_clear), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-close", G_CALLBACK (xpad_pad_close), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-undo", G_CALLBACK (xpad_pad_undo), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-redo", G_CALLBACK (xpad_pad_redo), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-cut", G_CALLBACK (xpad_pad_cut), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-copy", G_CALLBACK (xpad_pad_copy), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-paste", G_CALLBACK (xpad_pad_paste), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-delete", G_CALLBACK (xpad_pad_delete), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-properties", G_CALLBACK (xpad_pad_open_properties), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-preferences", G_CALLBACK (xpad_pad_open_preferences), pad);
	g_signal_connect_swapped (pad->priv->toolbar, "activate-quit", G_CALLBACK (xpad_pad_close_all), pad);

	g_signal_connect (pad->priv->toolbar, "popup", G_CALLBACK (xpad_pad_toolbar_popup), pad);
	g_signal_connect (pad->priv->toolbar, "popdown", G_CALLBACK (xpad_pad_toolbar_popdown), pad);

	g_signal_connect (pad->priv->menu, "deactivate", G_CALLBACK (xpad_pad_popup_deactivate), pad);
	g_signal_connect (pad->priv->highlight_menu, "deactivate", G_CALLBACK (xpad_pad_popup_deactivate), pad);
}

static void
xpad_pad_dispose (GObject *object)
{
	XpadPad *pad = XPAD_PAD (object);

	if (pad->priv->group) {
		g_object_unref(pad->priv->group);
		pad->priv->group = NULL;
	}

	if (GTK_IS_WIDGET(pad->priv->menu)) {
		gtk_widget_destroy (pad->priv->menu);
		pad->priv->menu = NULL;
	}

	if (GTK_IS_WIDGET(pad->priv->highlight_menu)) {
		gtk_widget_destroy (pad->priv->highlight_menu);
		pad->priv->highlight_menu = NULL;
	}

	/* For some reason the clipboard handler does not get automatically disconnected (or not at the right moment), leading to errors after deleting a pad. This manual disconnect prevents this error. */
	if (GTK_IS_CLIPBOARD(pad->priv->clipboard)) {
		g_signal_handlers_disconnect_matched (pad->priv->clipboard, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, pad);
	}

	/* For some reason the toolbar handler does not get automatically disconnected (or not at the right moment), leading to errors after deleting a pad. This manual disconnect prevents this error. */
	if (XPAD_IS_TOOLBAR(pad->priv->toolbar)) {
		g_signal_handlers_disconnect_matched (pad->priv->toolbar, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, pad);
		gtk_widget_destroy(pad->priv->toolbar);
		pad->priv->toolbar = NULL;
	}
	
	G_OBJECT_CLASS (xpad_pad_parent_class)->dispose (object);
}

static void
xpad_pad_finalize (GObject *object)
{
	XpadPad *pad = XPAD_PAD (object);

	if (xpad_global_settings)
		g_signal_handlers_disconnect_matched (xpad_global_settings, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, pad);

	g_free (pad->priv->infoname);
	g_free (pad->priv->contentname);
	
	G_OBJECT_CLASS (xpad_pad_parent_class)->finalize (object);
}

static void
xpad_pad_show (XpadPad *pad)
{
	/* Some wm's might not acknowledge our request for a specific
		location before we are shown.  What we do here is a little gimpy
		and not very respectful of wms' sovereignty, but it has the effect
		of making pads' locations very dependable.  We just move the pad
		again here after being shown.  This may create a visual effect if
		the wm did ignore us, but is better than being in the wrong
		place, I guess. */
	if (pad->priv->location_valid)
		gtk_window_move (GTK_WINDOW (pad), pad->priv->x, pad->priv->y);

	if (pad->priv->sticky)
		gtk_window_stick (GTK_WINDOW (pad));
	else
		gtk_window_unstick (GTK_WINDOW (pad));
}

static void
xpad_pad_notify_has_scrollbar (XpadPad *pad)
{
	gboolean has_scrollbar;
	g_object_get (xpad_global_settings, "has-scrollbar", &has_scrollbar, NULL);

	if (has_scrollbar)
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (pad->priv->scrollbar), 
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	else
	{
		GtkAdjustment *v, *h;
		
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (pad->priv->scrollbar), 
			GTK_POLICY_NEVER, GTK_POLICY_NEVER);
		
		/* now we need to adjust view so that user can see whole pad */
		h = gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (pad->priv->scrollbar));
		v = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (pad->priv->scrollbar));
		
		gtk_adjustment_set_value (h, 0);
		gtk_adjustment_set_value (v, 0);
	}
}

static void
xpad_pad_notify_has_decorations (XpadPad *pad)
{
	gboolean decorations;
	g_object_get (xpad_global_settings, "has-decorations", &decorations, NULL);
	
	/**
	 *  There are two modes of operation:  a normal mode and a 'stealth' mode.
	 *  If decorations are disabled, we also don't show up in the taskbar or pager. 
	 */
	gtk_window_set_decorated (GTK_WINDOW (pad), decorations);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW (pad), !decorations);
	gtk_window_set_skip_pager_hint (GTK_WINDOW (pad), !decorations);
	
	/* reshow_with_initial_size() seems to set the window back to a never-shown state.
		This is good, as some WMs don't like us changing the above parameters mid-run,
		even if we do a hide/show cycle. */
	gtk_window_set_default_size (GTK_WINDOW (pad), (gint) pad->priv->width, (gint) pad->priv->height);
	gtk_widget_hide(GTK_WIDGET (pad));
	gtk_widget_unrealize(GTK_WIDGET (pad));
	gtk_widget_show(GTK_WIDGET (pad));
}

static guint
xpad_pad_text_and_toolbar_height (XpadPad *pad)
{
	cairo_rectangle_int_t rec;
	gint textx, texty, x, y;
	GtkTextIter iter;
	
	gtk_text_view_get_visible_rect(GTK_TEXT_VIEW(pad->priv->textview), &rec);
	gtk_text_buffer_get_end_iter(gtk_text_view_get_buffer(GTK_TEXT_VIEW(pad->priv->textview)), &iter);
	gtk_text_view_get_iter_location(GTK_TEXT_VIEW(pad->priv->textview), &iter, &rec);
	gtk_text_view_buffer_to_window_coords(GTK_TEXT_VIEW(pad->priv->textview),
		GTK_TEXT_WINDOW_WIDGET, rec.x + rec.width, rec.y + rec.height,
		&textx, &texty);
	gtk_widget_translate_coordinates(pad->priv->textview, GTK_WIDGET(pad), textx, texty, &x, &y);

	/* Safe cast from gint to guint */
	if (y >= 0) {
		return (guint) y + pad->priv->toolbar_height + gtk_container_get_border_width(GTK_CONTAINER(pad->priv->textview));
	}
	else {
		g_warning("There is a problem in the program Xpad. In function 'xpad_pad_toolbar_size_allocate' the variable 'event->height' is not a postive number. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
		return 0;
	}
}

static void
xpad_pad_show_toolbar (XpadPad *pad)
{
	if (!gtk_widget_get_visible (pad->priv->toolbar))
	{
		GtkRequisition req;
		
		if (gtk_widget_get_window(GTK_WIDGET(pad)))
			gdk_window_freeze_updates (gtk_widget_get_window(GTK_WIDGET(pad)));
		gtk_widget_show (pad->priv->toolbar);
		if (!pad->priv->toolbar_height)
		{
			gtk_widget_get_preferred_size (pad->priv->toolbar, &req, NULL);
			/* safe cast from gint to guint */
			if (req.height >= 0) {
				pad->priv->toolbar_height = (guint) req.height;
			}
			else {
				g_warning ("There is a problem in the program Xpad. In function 'xpad_pad_show_toolbar' the variable 'req.height' is not a postive number. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
				pad->priv->toolbar_height = 0;
			}
		}

		/* Do we have room for the toolbar without covering text? */
		if (xpad_pad_text_and_toolbar_height (pad) > pad->priv->height)
		{
			pad->priv->toolbar_expanded = TRUE;
			pad->priv->height += pad->priv->toolbar_height;
			gtk_window_resize (GTK_WINDOW (pad), (gint) pad->priv->width, (gint) pad->priv->height);
		}
		else
			pad->priv->toolbar_expanded = FALSE;
		
		pad->priv->toolbar_pad_resized = FALSE;
		
		if (gtk_widget_get_window(GTK_WIDGET(pad)))
			gdk_window_thaw_updates (gtk_widget_get_window(GTK_WIDGET(pad)));
	}
}

static void
xpad_pad_hide_toolbar (XpadPad *pad)
{
	if (gtk_widget_get_visible (pad->priv->toolbar))
	{
		if (gtk_widget_get_window(GTK_WIDGET(pad)))
			gdk_window_freeze_updates (gtk_widget_get_window(GTK_WIDGET(pad)));
		gtk_widget_hide (pad->priv->toolbar);
		
		if (pad->priv->toolbar_expanded ||
			 (pad->priv->toolbar_pad_resized && xpad_pad_text_and_toolbar_height (pad) >= pad->priv->height))
		{
				pad->priv->height -= pad->priv->toolbar_height;
				gtk_window_resize (GTK_WINDOW (pad), (gint) pad->priv->width, (gint) pad->priv->height);
				pad->priv->toolbar_expanded = FALSE;
		}
		if (gtk_widget_get_window(GTK_WIDGET(pad)))
			gdk_window_thaw_updates (gtk_widget_get_window(GTK_WIDGET(pad)));
	}
}

static void
xpad_pad_notify_has_toolbar (XpadPad *pad)
{
	gboolean has_toolbar, autohide_toolbar;
	g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, "autohide-toolbar", &autohide_toolbar, NULL);

	if (has_toolbar && !autohide_toolbar)
		xpad_pad_show_toolbar (pad);
	else
		xpad_pad_hide_toolbar (pad);
}

static gboolean
toolbar_timeout (XpadPad *pad)
{
	if (!pad || !pad->priv || !pad->priv->toolbar_timeout)
		return FALSE;

	gboolean has_toolbar, autohide_toolbar;
	g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, "autohide-toolbar", &autohide_toolbar, NULL);

	if (pad->priv->toolbar_timeout && autohide_toolbar && has_toolbar)
		xpad_pad_hide_toolbar (pad);
	
	pad->priv->toolbar_timeout = 0;
	
	return FALSE;
}

static void
xpad_pad_notify_autohide_toolbar (XpadPad *pad)
{
	gboolean autohide_toolbar;
	g_object_get (xpad_global_settings, "autohide-toolbar", &autohide_toolbar, NULL);

	if (autohide_toolbar)
	{
		/* Likely not to be in pad when turning setting on */
		if (!pad->priv->toolbar_timeout)
			pad->priv->toolbar_timeout = g_timeout_add (1000, (GSourceFunc) toolbar_timeout, pad);
	}
	else
	{
		gboolean has_toolbar;
		g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, NULL);

		if (has_toolbar)
			xpad_pad_show_toolbar(pad);
	}
}

void
xpad_pad_notify_has_selection (XpadPad *pad)
{
	g_return_if_fail (pad);

	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	gboolean has_selection = gtk_text_buffer_get_has_selection (buffer);

	XpadToolbar *toolbar = XPAD_TOOLBAR (pad->priv->toolbar);
	if (toolbar == NULL)
		return;

	xpad_toolbar_enable_cut_button (toolbar, has_selection);
	xpad_toolbar_enable_copy_button (toolbar, has_selection);
}

void
xpad_pad_notify_clipboard_owner_changed (XpadPad *pad)
{
	g_return_if_fail (pad);

	XpadToolbar *toolbar = NULL;
	/* safe cast to toolbar */
	if (XPAD_IS_TOOLBAR (pad->priv->toolbar)) {
		toolbar = XPAD_TOOLBAR (pad->priv->toolbar);
		g_return_if_fail (toolbar);

		GtkClipboard *clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
		xpad_toolbar_enable_paste_button (toolbar, gtk_clipboard_wait_is_text_available (clipboard));
	}
	else {
		g_warning("There is a problem in the program Xpad. In function 'xpad_pad_notify_clipboard_owner_changed' the variable 'pad->priv->toolbar' is not of type toolbar. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
	}
}

void
xpad_pad_notify_undo_redo_changed (XpadPad *pad)
{
	g_return_if_fail (pad);

	XpadTextBuffer *buffer = NULL;
	buffer = XPAD_TEXT_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)));
	g_return_if_fail (buffer);

	XpadToolbar *toolbar = NULL;
	toolbar = XPAD_TOOLBAR (pad->priv->toolbar);
	g_return_if_fail (toolbar);

	xpad_toolbar_enable_undo_button (toolbar, xpad_text_buffer_undo_available (buffer));
	xpad_toolbar_enable_redo_button (toolbar, xpad_text_buffer_redo_available (buffer));
}

static gboolean
xpad_pad_enter_notify_event (GtkWidget *pad, GdkEventCrossing *event)
{
	gboolean has_toolbar, autohide_toolbar;
	g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, "autohide-toolbar", &autohide_toolbar, NULL);

	if (has_toolbar && autohide_toolbar &&
		 event->detail != GDK_NOTIFY_INFERIOR &&
		 event->mode == GDK_CROSSING_NORMAL)
	{
		XPAD_PAD (pad)->priv->toolbar_timeout = 0;
		xpad_pad_show_toolbar (XPAD_PAD (pad));
	}
	
	return FALSE;
}

static gboolean
xpad_pad_leave_notify_event (GtkWidget *pad, GdkEventCrossing *event)
{
	gboolean has_toolbar, autohide_toolbar;
	g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, "autohide-toolbar", &autohide_toolbar, NULL);

	if (has_toolbar && autohide_toolbar &&
		 event->detail != GDK_NOTIFY_INFERIOR &&
		 event->mode == GDK_CROSSING_NORMAL)
	{
		if (!XPAD_PAD (pad)->priv->toolbar_timeout)
			XPAD_PAD (pad)->priv->toolbar_timeout = g_timeout_add (1000, (GSourceFunc) toolbar_timeout, pad);
	}
	
	return FALSE;
}

static void
xpad_pad_spawn (XpadPad *pad)
{
	GtkWidget *newpad = xpad_pad_new (pad->priv->group);
	gtk_widget_show (newpad);
}

static void
xpad_pad_clear (XpadPad *pad)
{
	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	gtk_text_buffer_set_text (buffer, "", -1);
}

void
xpad_pad_close (XpadPad *pad)
{
	gtk_widget_hide (GTK_WIDGET (pad));
	
	/* If no tray and this is the last pad, we don't want to record this
		pad as closed, we want to start with just this pad next open.  So
		quit before we record. */
	if (!xpad_tray_is_open () &&
		 xpad_pad_group_num_visible_pads (pad->priv->group) == 0)
	{
		xpad_pad_quit (pad);
		return;
	}
	
	if (pad->priv->properties)
		gtk_widget_destroy (pad->priv->properties);
	
	xpad_pad_save_info (pad);
	
	g_signal_emit (pad, signals[CLOSED], 0);
}

void
xpad_pad_toggle(XpadPad *pad)
{
	 if (gtk_widget_get_visible (GTK_WIDGET(pad)))
		  xpad_pad_close (pad);
	 else
		  gtk_widget_show (GTK_WIDGET (pad));
}

static gboolean
should_confirm_delete (XpadPad *pad)
{
	GtkTextBuffer *buffer;
	GtkTextIter s, e;
	gchar *content;
	gboolean confirm;
	
	g_object_get (xpad_global_settings, "confirm-destroy", &confirm, NULL);
	if (!confirm)
		return FALSE;
	
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	gtk_text_buffer_get_bounds (buffer, &s, &e);
	content = gtk_text_buffer_get_text (buffer, &s, &e, FALSE);
	
	confirm = strcmp (g_strstrip (content), "") != 0;
	
	g_free (content);
	
	return confirm;
}

static void
xpad_pad_delete (XpadPad *pad)
{
	g_return_if_fail (pad);

	/* With the delayed saving functionality, it is necessary to clear the unsaved flags to prevent usage of non-existing object information. */
	pad->priv->unsaved_info = FALSE;
	pad->priv->unsaved_content = FALSE;

	if (should_confirm_delete (pad))
	{
		GtkWidget *dialog;
		gint response;

		dialog = xpad_app_alert_dialog (GTK_WINDOW (pad), "dialog-warning", _("Delete this pad?"), _("All text of this pad will be irrevocably lost."));
		
		if (!dialog)
			return;

		gtk_dialog_add_buttons (GTK_DIALOG (dialog), _("_Delete"), GTK_RESPONSE_ACCEPT, _("_Cancel"), GTK_RESPONSE_REJECT, NULL);

		response = gtk_dialog_run (GTK_DIALOG (dialog));
		
		gtk_widget_destroy (dialog);
		
		if (response != GTK_RESPONSE_ACCEPT)
			return;
	}
	
	/* These two if statements actually erase the pad on the harddisk. */
	if (pad->priv->infoname)
		fio_remove_file (pad->priv->infoname);
	if (pad->priv->contentname)
		fio_remove_file (pad->priv->contentname);
	
	/* Remove the pad from the group and destroy it. */
	gtk_widget_destroy (GTK_WIDGET (pad));
}

static void
pad_properties_sync_title (XpadPad *pad)
{
	gchar *title;
	
	if (!pad->priv->properties)
		return;
	
	title = g_strdup_printf (_("'%s' Properties"), gtk_window_get_title (GTK_WINDOW (pad)));
	gtk_window_set_title (GTK_WINDOW (pad->priv->properties), title);
	g_free (title);
}

static void
pad_properties_destroyed (XpadPad *pad)
{
	if (!pad->priv->properties)
		return;
	
	g_signal_handlers_disconnect_by_func (pad, (gpointer) pad_properties_sync_title, NULL);
	pad->priv->properties = NULL;
}

static void
prop_notify_follow_font (XpadPad *pad)
{
	XpadPadProperties *prop = XPAD_PAD_PROPERTIES (pad->priv->properties);

	gboolean follow_font_style;
	g_object_get (prop, "follow-font-style", &follow_font_style, NULL);
	xpad_text_view_set_follow_font_style (XPAD_TEXT_VIEW (pad->priv->textview), follow_font_style);
	
	if (!follow_font_style)
	{
		const gchar *font;
		g_object_get (prop, "fontname", &font, NULL);
		PangoFontDescription *fontdesc;
		
		fontdesc = font ? pango_font_description_from_string (font) : NULL;
		gtk_widget_override_font (pad->priv->textview, fontdesc);
		if (fontdesc)
			pango_font_description_free (fontdesc);
	}
	
	xpad_pad_save_info_delayed (pad);
}

static void
prop_notify_colors (XpadPad *pad)
{
	XpadPadProperties *prop = XPAD_PAD_PROPERTIES (pad->priv->properties);
	
	gboolean follow_color_style;
	g_object_get (prop, "follow-color-style", &follow_color_style, NULL);
	xpad_text_view_set_follow_color_style (XPAD_TEXT_VIEW (pad->priv->textview), follow_color_style);

	if (follow_color_style)
	{
		/* Set the colors to the global preferences colors */
		const GdkRGBA *text_color, *back_color;
		g_object_get (xpad_global_settings, "text-color", &text_color, "back-color", &back_color, NULL);

		gtk_widget_override_cursor (pad->priv->textview, text_color, text_color);
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, text_color);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, back_color);

		/* Inverse the text and background colors for selected text, so it is likely to be visible by any choice of the colors. */
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, back_color);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, text_color);
	}
	else
	{
		/* Set the color to the individual pad properties colors */
		const GdkRGBA *text_color, *back_color;
		g_object_get (prop, "text-color", &text_color, "back-color", &back_color, NULL);

		gtk_widget_override_cursor (pad->priv->textview, text_color, text_color);
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, text_color);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, back_color);

		/* Inverse the text and background colors for selected text, so it is likely to be visible by any choice of the colors. */
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, back_color);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, text_color);
	}

	xpad_pad_save_info_delayed (pad);
}

static void
prop_notify_font (XpadPad *pad)
{
	XpadPadProperties *prop = XPAD_PAD_PROPERTIES (pad->priv->properties);
	
	const gchar *font;
	g_object_get (prop, "fontname", &font, NULL);
	PangoFontDescription *fontdesc;
	
	fontdesc = font ? pango_font_description_from_string (font) : NULL;
	gtk_widget_override_font (pad->priv->textview, fontdesc);
	if (fontdesc)
		pango_font_description_free (fontdesc);
	
	xpad_pad_save_info_delayed (pad);
}

static void
xpad_pad_open_properties (XpadPad *pad)
{
	GtkStyleContext *style = NULL;
	PangoFontDescription *font;
	GdkRGBA widget_text_color = {0, 0, 0, 0};
	GdkRGBA widget_background_color = {0, 0, 0, 0};
	
	if (pad->priv->properties)
	{
		gtk_window_present (GTK_WINDOW (pad->priv->properties));
		return;
	}
	
	pad->priv->properties = xpad_pad_properties_new ();
	
	gtk_window_set_transient_for (GTK_WINDOW (pad->priv->properties), GTK_WINDOW (pad));
	gtk_window_set_resizable (GTK_WINDOW (pad->priv->properties), FALSE);
	
	g_signal_connect_swapped (pad->priv->properties, "destroy", G_CALLBACK (pad_properties_destroyed), pad);
	g_signal_connect (pad, "notify::title", G_CALLBACK (pad_properties_sync_title), NULL);

	style = gtk_widget_get_style_context (pad->priv->textview);
	gtk_style_context_get(style, GTK_STATE_FLAG_NORMAL, GTK_STYLE_PROPERTY_FONT, &font, NULL);
	gtk_style_context_get_color (style, GTK_STATE_FLAG_NORMAL, &widget_text_color);
	gtk_style_context_get_background_color (style, GTK_STATE_FLAG_NORMAL, &widget_background_color);

	g_object_set (G_OBJECT (pad->priv->properties),
		"follow-font-style", xpad_text_view_get_follow_font_style (XPAD_TEXT_VIEW (pad->priv->textview)),
		"follow-color-style", xpad_text_view_get_follow_color_style (XPAD_TEXT_VIEW (pad->priv->textview)),
		"text-color", &widget_text_color,
		"back-color", &widget_background_color,
		"fontname", pango_font_description_to_string(font),
		NULL);
	pango_font_description_free (font);
	
	g_signal_connect_swapped (pad->priv->properties, "notify::follow-font-style", G_CALLBACK (prop_notify_follow_font), pad);
	g_signal_connect_swapped (pad->priv->properties, "notify::follow-color-style", G_CALLBACK (prop_notify_colors), pad);
	g_signal_connect_swapped (pad->priv->properties, "notify::text-color", G_CALLBACK (prop_notify_colors), pad);
	g_signal_connect_swapped (pad->priv->properties, "notify::back-color", G_CALLBACK (prop_notify_colors), pad);
	g_signal_connect_swapped (pad->priv->properties, "notify::fontname", G_CALLBACK (prop_notify_font), pad);
	
	pad_properties_sync_title (pad);
	
	gtk_widget_show (pad->priv->properties);
}

static void
xpad_pad_open_preferences ()
{
	xpad_preferences_open ();
}

static void
xpad_pad_quit ()
{
	xpad_app_quit ();
}

static void
xpad_pad_text_changed (XpadPad *pad, GtkTextBuffer *buffer)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) buffer;

	/* set title */
	xpad_pad_sync_title (pad);
	
	/* record change */
	xpad_pad_save_content_delayed(pad);
}

static gboolean
xpad_pad_toolbar_size_allocate (XpadPad *pad, GtkAllocation *event)
{
	/* safe cast from gint to guint */
	if (event->height >= 0) {
		pad->priv->toolbar_height = (guint) event->height;
	}
	else {
		g_warning("There is a problem in the program Xpad. In function 'xpad_pad_toolbar_size_allocate' the variable 'event->height' is not a postive number. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
		pad->priv->toolbar_height = 0;
	}
	return FALSE;
}

static gboolean
xpad_pad_configure_event (XpadPad *pad, GdkEventConfigure *event)
{
	if (!gtk_widget_get_visible (GTK_WIDGET(pad)))
		return FALSE;

	int eWidth = event->width;
	int eHeight = event->height;
	
	/* safe cast from gint to guint */
	if (eWidth >= 0 && eHeight >=0 ) {
		if (pad->priv->width != (guint) eWidth || pad->priv->height != (guint) eHeight)
			pad->priv->toolbar_pad_resized = TRUE;

		pad->priv->width = (guint) event->width;
		pad->priv->height = (guint) event->height;
	}
	else {
		g_warning("There is a problem in the program Xpad. In function 'xpad_pad_configure_event' the variable 'event->width' or 'event->height' is not a postive number. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
	}
	
	pad->priv->x = event->x;
	pad->priv->y = event->y;
	pad->priv->location_valid = TRUE;
	
	xpad_pad_save_info_delayed(pad);
	
	/* Sometimes when moving, if the toolbar tries to hide itself,
		the window manager will not resize it correctly.  So, we make
		sure not to end the timeout while moving. */
	if (pad->priv->toolbar_timeout)
	{
		g_source_remove (pad->priv->toolbar_timeout);
		pad->priv->toolbar_timeout = g_timeout_add (1000, (GSourceFunc) toolbar_timeout, pad);
	}
	
	return FALSE;
}

static gboolean
xpad_pad_delete_event (XpadPad *pad, GdkEvent *event)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) event;

	xpad_pad_close (pad);
	
	return TRUE;
}

static gboolean
xpad_pad_popup_menu (XpadPad *pad)
{
	xpad_pad_popup (pad, NULL);
	
	return TRUE;
}

static gboolean
xpad_pad_text_view_button_press_event (GtkWidget *text_view, GdkEventButton *event, XpadPad *pad)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) text_view;

	if (event->type == GDK_BUTTON_PRESS)
	{
		switch (event->button)
		{
		case 1:
			if ((event->state & gtk_accelerator_get_default_mod_mask ()) == GDK_CONTROL_MASK)
			{
				gtk_window_begin_move_drag (GTK_WINDOW (pad), (gint) event->button, (gint) event->x_root, (gint) event->y_root, event->time);
				return TRUE;
			}
			break;
		
		case 3:
			if ((event->state & gtk_accelerator_get_default_mod_mask ()) == GDK_CONTROL_MASK)
			{
				GdkWindowEdge edge;
				
				if (gtk_widget_get_direction (GTK_WIDGET (pad)) == GTK_TEXT_DIR_LTR)
					edge = GDK_WINDOW_EDGE_SOUTH_EAST;
				else
					edge = GDK_WINDOW_EDGE_SOUTH_WEST;
				
				gtk_window_begin_resize_drag (GTK_WINDOW (pad), edge, (gint) event->button, (gint) event->x_root, (gint) event->y_root, event->time);
			}
			else
			{
				xpad_pad_popup (pad, event);
			}
			return TRUE;
		}
	}
	
	return FALSE;
}

static gboolean
xpad_pad_button_press_event (XpadPad *pad, GdkEventButton *event)
{
	if (event->type == GDK_BUTTON_PRESS)
	{
		switch (event->button)
		{
		case 1:
			gtk_window_begin_move_drag (GTK_WINDOW (pad), (gint) event->button, (gint) event->x_root, (gint) event->y_root, event->time);
			return TRUE;
		
		case 3:
			if ((event->state & gtk_accelerator_get_default_mod_mask ()) == GDK_CONTROL_MASK)
			{
				GdkWindowEdge edge;
				
				if (gtk_widget_get_direction (GTK_WIDGET (pad)) == GTK_TEXT_DIR_LTR)
					edge = GDK_WINDOW_EDGE_SOUTH_EAST;
				else
					edge = GDK_WINDOW_EDGE_SOUTH_WEST;
				
				gtk_window_begin_resize_drag (GTK_WINDOW (pad), edge, (gint) event->button, (gint) event->x_root, (gint) event->y_root, event->time);
			}
			else
			{
				xpad_pad_popup (pad, event);
			}
			return TRUE;
		}
	}
	
	return FALSE;
}

static void
xpad_pad_sync_title (XpadPad *pad)
{
	GtkTextBuffer *buffer;
	GtkTextIter s, e;
	gchar *content, *end;
	
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	gtk_text_buffer_get_bounds (buffer, &s, &e);
	content = gtk_text_buffer_get_text (buffer, &s, &e, FALSE);
	end = g_utf8_strchr (content, -1, '\n');
	if (end)
		*end = '\0';
	
	gtk_window_set_title (GTK_WINDOW (pad), g_strstrip (content));
	
	g_free (content);
}

static void
xpad_pad_set_group (XpadPad *pad, XpadPadGroup *group)
{
	pad->priv->group = group;
	g_object_ref(pad->priv->group);

	if (group)
		xpad_pad_group_add (group, GTK_WIDGET (pad));
}

static XpadPadGroup *
xpad_pad_get_group (XpadPad *pad)
{
	return pad->priv->group;
}

static void
xpad_pad_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	XpadPad *pad;
	
	pad = XPAD_PAD (object);
	
	switch (prop_id)
	{
	case PROP_GROUP:
		xpad_pad_set_group (pad, g_value_get_pointer (value));
		break;
	
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
xpad_pad_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	XpadPad *pad;
	
	pad = XPAD_PAD (object);
	
	switch (prop_id)
	{
	case PROP_GROUP:
		g_value_set_pointer (value, xpad_pad_get_group (pad));
		break;
	
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

void
xpad_pad_load_content (XpadPad *pad)
{
	g_return_if_fail (pad);

	gchar *content;
	GtkTextBuffer *buffer;
	
	if (!pad->priv->contentname)
		return;
	
	content = fio_get_file (pad->priv->contentname);
	
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	
	xpad_text_buffer_freeze_undo (XPAD_TEXT_BUFFER (buffer));
	g_signal_handlers_block_by_func (buffer, xpad_pad_text_changed, pad);
	
	xpad_text_buffer_set_text_with_tags (XPAD_TEXT_BUFFER (buffer), content ? content : "");
	g_free (content);
	
	g_signal_handlers_unblock_by_func (buffer, xpad_pad_text_changed, pad);
	xpad_text_buffer_thaw_undo (XPAD_TEXT_BUFFER (buffer));

	xpad_pad_text_changed(pad, buffer);
	pad->priv->unsaved_content = FALSE;
}

void
xpad_pad_save_content (XpadPad *pad)
{
	g_return_if_fail (pad);

	gchar *content;
	GtkTextBuffer *buffer;
	
	if (!pad->priv->unsaved_content)
		return;

	/* create content file if it doesn't exist yet */
	if (!pad->priv->contentname)
	{
		pad->priv->contentname = fio_unique_name ("content-");
		if (!pad->priv->contentname)
			return;
	}

	if (GTK_IS_TEXT_VIEW(pad->priv->textview)) {
		buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
		content = xpad_text_buffer_get_text_with_tags (XPAD_TEXT_BUFFER (buffer));
	}
	else
		g_warning("There is a problem in the program Xpad. In function 'xpad_pad_save_content' the variable 'pad->priv->textview' is not of type textview. Please send a bugreport to https://bugs.launchpad.net/xpad/+filebug to help improve Xpad.");
	
	fio_set_file (pad->priv->contentname, content);

	pad->priv->unsaved_content = FALSE;
	g_free (content);
}

/* Extract all the metadata of a single pad from its info-xxxxx file and store it in the pad object */
static void
xpad_pad_load_info (XpadPad *pad, gboolean *show)
{
	gboolean locked = FALSE, follow_font = TRUE, follow_color = TRUE;
	gboolean hidden = FALSE;
	gchar *text_color_string = NULL;
	gchar *background_color_string = NULL;
	GdkRGBA text = {0, 0, 0, 0};
	GdkRGBA back = {0, 0, 0, 0};
	gchar *fontname = NULL;

	if (!pad->priv->infoname)
		return;
	
	if (fio_get_values_from_file (pad->priv->infoname, 
		"i|width", &pad->priv->width,
		"i|height", &pad->priv->height,
		"i|x", &pad->priv->x,
		"i|y", &pad->priv->y,
		"b|locked", &locked,
		"b|follow_font", &follow_font,
		"b|follow_color", &follow_color,
		"b|sticky", &pad->priv->sticky,
		"b|hidden", &hidden,
		"s|back", &background_color_string,
		"s|text", &text_color_string,
		"s|fontname", &fontname,
		"s|content", &pad->priv->contentname,
		NULL))
		return;

	pad->priv->unsaved_info = FALSE;
	
	pad->priv->location_valid = TRUE;

	gboolean has_toolbar, autohide_toolbar;
	g_object_get (xpad_global_settings, "has-toolbar", &has_toolbar, "autohide-toolbar", &autohide_toolbar, NULL);

	if (has_toolbar && !autohide_toolbar)
	{
		pad->priv->toolbar_height = 0;
		xpad_pad_hide_toolbar (pad);
		xpad_pad_show_toolbar (pad); /* these will resize pad at correct height */
	}
	else
		gtk_window_resize (GTK_WINDOW (pad), (gint) pad->priv->width, (gint) pad->priv->height);
	gtk_window_move (GTK_WINDOW (pad), pad->priv->x, pad->priv->y);
	
	xpad_text_view_set_follow_font_style (XPAD_TEXT_VIEW (pad->priv->textview), follow_font);
	xpad_text_view_set_follow_color_style (XPAD_TEXT_VIEW (pad->priv->textview), follow_color);
	
	if (locked)
	{
		xpad_text_view_set_follow_font_style (XPAD_TEXT_VIEW (pad->priv->textview), FALSE);
		xpad_text_view_set_follow_color_style (XPAD_TEXT_VIEW (pad->priv->textview), FALSE);
	}
	
	if (!follow_font)
	{
		PangoFontDescription *font_desc = pango_font_description_from_string (fontname);
		gtk_widget_override_font (pad->priv->textview, font_desc);
		pango_font_description_free (font_desc);
	}

	if (!follow_color)
	{
		/*
		 * If, for some reason, one of the colors could not be retrieved
		 * (for example due to the migration to the new GdkRGBA colors),
		 * set the color to the default.
		 */
		if (text_color_string == NULL || background_color_string == NULL) {
			text = (GdkRGBA) {0, 0, 0, 1};
			back = (GdkRGBA) {1, 0.933334350586, 0.6, 1};
		}
		else {
			/* If, for some reason, the parsing of the colors fail, set the color to the default. */
			if (!gdk_rgba_parse (&text, text_color_string) || !gdk_rgba_parse (&back, background_color_string)) {
				text = (GdkRGBA) {0, 0, 0, 1};
				back = (GdkRGBA) {1, 0.933334350586, 0.6, 1};
			}
		}

		/* Set the text and background color for this pad, as stated in its properties file. */
		gtk_widget_override_cursor (pad->priv->textview, &text, &text);
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, &text);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_NORMAL, &back);

		/* Inverse the text and background colors for selected text, so it is likely to be visible by any choice of the colors. */
		gtk_widget_override_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, &back);
		gtk_widget_override_background_color (pad->priv->textview, GTK_STATE_FLAG_SELECTED, &text);
	}
	
	/*
	 * Find the sticky notes menu setting for this pad (which is on the global default),
	 * and change its setting to the setting from the info file (pad specific default).
	 */
	if(GTK_IS_CONTAINER(pad->priv->menu)) {
		GObject *obj;
		GList *elem, *children;
		children = gtk_container_get_children(GTK_CONTAINER(pad->priv->menu));

		for(elem = children; elem; elem = elem->next) {
			obj = (GObject *) elem->data;

			if(GTK_IS_BIN(obj) && GTK_IS_MENU_ITEM(obj)) {
				GList *elem2, *children2;
				children2 = gtk_container_get_children(GTK_CONTAINER(gtk_menu_item_get_submenu(GTK_MENU_ITEM(obj))));
				for(elem2 = children2; elem2; elem2 = elem2->next) {
					obj = (GObject *) elem2->data;
					if (GTK_IS_CHECK_MENU_ITEM(obj)) {
						if (!g_strcmp0(gtk_menu_item_get_label(GTK_MENU_ITEM(obj)), "Show on _All Workspaces")) {
							gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(obj), pad->priv->sticky);
							break;
						}
					}
				}
				g_list_free(elem2);
			}
		}
		g_list_free(elem);
	}
	else
		g_warning("For some reason the variable pad->priv->menu is not a container. This results in not having a unique sticky setting for this pad. Falling back to the global sticky setting. Please send a bugreport.");

	if (pad->priv->sticky)
		gtk_window_stick (GTK_WINDOW (pad));
	else
		gtk_window_unstick (GTK_WINDOW (pad));
	
	if (show)
		*show = !hidden;

	g_free(fontname);
}

void
xpad_pad_save_info (XpadPad *pad)
{
	guint height = NULL;
	GtkStyleContext *style = NULL;
	PangoFontDescription *font = NULL;
	GdkRGBA widget_text_color = {0, 0, 0, 0};
	GdkRGBA widget_background_color = {0, 0, 0, 0};

	g_return_if_fail (pad);

	if (!pad->priv->unsaved_info)
		return;
	
	/* Must create pad info file if it doesn't exist yet */
	if (!pad->priv->infoname)
	{
		pad->priv->infoname = fio_unique_name ("info-");
		if (!pad->priv->infoname)
			return;
		gtk_window_set_role (GTK_WINDOW (pad), pad->priv->infoname);
	}
	/* create content file if it doesn't exist yet */
	if (!pad->priv->contentname)
	{
		pad->priv->contentname = fio_unique_name ("content-");
		if (!pad->priv->contentname)
			return;
	}
	
	height = pad->priv->height;
	if (gtk_widget_get_visible (pad->priv->toolbar) && pad->priv->toolbar_expanded)
		height -= pad->priv->toolbar_height;
	
	style = gtk_widget_get_style_context (pad->priv->textview);
	gtk_style_context_get (style, GTK_STATE_FLAG_NORMAL, GTK_STYLE_PROPERTY_FONT, &font, NULL);
	gtk_style_context_get_color (style, GTK_STATE_FLAG_NORMAL, &widget_text_color);
	gtk_style_context_get_background_color (style, GTK_STATE_FLAG_NORMAL, &widget_background_color);

	fio_set_values_to_file (pad->priv->infoname,
		"i|width", pad->priv->width,
		"i|height", height,
		"i|x", pad->priv->x,
		"i|y", pad->priv->y,
		"b|follow_font", xpad_text_view_get_follow_font_style (XPAD_TEXT_VIEW (pad->priv->textview)),
		"b|follow_color", xpad_text_view_get_follow_color_style (XPAD_TEXT_VIEW (pad->priv->textview)),
		"b|sticky", pad->priv->sticky,
		"b|hidden", !gtk_widget_get_visible (GTK_WIDGET(pad)),
		"s|back", gdk_rgba_to_string (&widget_background_color),
		"s|text", gdk_rgba_to_string (&widget_text_color),
		"s|fontname", pango_font_description_to_string (font),
		"s|content", pad->priv->contentname,
		NULL);

	pango_font_description_free (font);
	pad->priv->unsaved_info = FALSE;
}

static void
menu_about (XpadPad *pad)
{
	const gchar *artists[] = {"Michael Terry <mike@mterry.name>", NULL};
	const gchar *authors[] = {"Arthur Borsboom <arthurborsboom@gmail.com>", "Jeroen Vermeulen <jtv@xs4all.nl>", "Michael Terry <mike@mterry.name>", "Paul Ivanov <pivanov@berkeley.edu>", "Sachin Raut <great.sachin@gmail.com>", NULL};
	const gchar *comments = _("Sticky notes");
	const gchar *copyright = "© 2001-2014 Michael Terry";
	/* Translators: please translate this as your own name and optionally email
		like so: "Your Name <your@email.com>" */
	const gchar *translator_credits = _("translator-credits");
	const gchar *website = "https://launchpad.net/xpad";
	
	gtk_show_about_dialog (GTK_WINDOW (pad),
		"artists", artists,
		"authors", authors,
		"comments", comments,
		"copyright", copyright,
		"license-type", GTK_LICENSE_GPL_3_0,
		"logo-icon-name", PACKAGE,
		"translator-credits", translator_credits,
		"version", VERSION,
		"website", website,
		NULL);
}

void
xpad_pad_cut (XpadPad *pad)
{
	gtk_text_buffer_cut_clipboard (
		gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)),
		gtk_clipboard_get (GDK_SELECTION_CLIPBOARD),
		TRUE);
}

static void
menu_cut (XpadPad *pad)
{
	xpad_pad_cut (pad);
}

void
xpad_pad_copy (XpadPad *pad)
{
	gtk_text_buffer_copy_clipboard (
		gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)),
		gtk_clipboard_get (GDK_SELECTION_CLIPBOARD));
}

static void
menu_copy (XpadPad *pad)
{
	xpad_pad_copy (pad);
}

void
xpad_pad_paste (XpadPad *pad)
{
	gtk_text_buffer_paste_clipboard (
		gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)),
		gtk_clipboard_get (GDK_SELECTION_CLIPBOARD),
		NULL,
		TRUE);
}

static void
menu_paste (XpadPad *pad)
{
	xpad_pad_paste (pad);
}

void
xpad_pad_undo (XpadPad *pad)
{
	g_return_if_fail (pad->priv->textview);
	XpadTextBuffer *buffer = NULL;
	buffer = XPAD_TEXT_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)));
	g_return_if_fail (buffer);
	xpad_text_buffer_undo (buffer);
}

static void
menu_undo (XpadPad *pad)
{
	xpad_pad_undo (pad);
}

void
xpad_pad_redo (XpadPad *pad)
{
	g_return_if_fail (pad->priv->textview);
	XpadTextBuffer *buffer = NULL;
	buffer = XPAD_TEXT_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)));
	g_return_if_fail (buffer);
	xpad_text_buffer_redo (buffer);
}

static void
menu_redo (XpadPad *pad)
{
	xpad_pad_redo (pad);
}

static void
menu_show_all (XpadPad *pad)
{
	GSList *pads, *i;
	
	if (!pad->priv->group)
		return;
	
	pads = xpad_pad_group_get_pads (pad->priv->group);
	
	for (i = pads; i; i = i->next)
	{
		if (XPAD_PAD (i->data) != pad)
			gtk_window_present (GTK_WINDOW (i->data));
	}
	gtk_window_present (GTK_WINDOW (pad));
	
	g_slist_free (pads);
	g_slist_free (i);
	pads = NULL;
	i = NULL;
}

static void
xpad_pad_close_all (XpadPad *pad)
{
	if (!pad->priv->group)
		return;
	
	/**
	 * The logic is different here depending on whether the tray is open.
	 * If it is open, we just close each pad individually.  If it isn't
	 * open, we do a quit.  This way, when xpad is run again, only the
	 * pads open during the last 'close all' will open again.
	 */
	if (xpad_tray_is_open ())
		xpad_pad_group_close_all (pad->priv->group);
	else
		xpad_pad_quit (pad);
}

static void
menu_show (XpadPad *pad)
{
	gtk_window_present (GTK_WINDOW (pad));
}

static void
menu_toggle_tag (XpadPad *pad, const gchar *name)
{
	g_return_if_fail (pad->priv->textview);
	XpadTextBuffer *buffer = NULL;
	buffer = XPAD_TEXT_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview)));
	xpad_text_buffer_toggle_tag (buffer, name);
	xpad_pad_save_content_delayed(pad);
}

static void
menu_bold (XpadPad *pad)
{
	menu_toggle_tag (pad, "bold");
}

static void
menu_italic (XpadPad *pad)
{
	menu_toggle_tag (pad, "italic");
}

static void
menu_underline (XpadPad *pad)
{
	menu_toggle_tag (pad, "underline");
}

static void
menu_strikethrough (XpadPad *pad)
{
	menu_toggle_tag (pad, "strikethrough");
}

/* Make the pad visually stick to the workspace and save this setting to the individual pad info file,
 * because this function has been probably been called, because of a menu toggle. */
static void
menu_sticky (XpadPad *pad, GtkCheckMenuItem *check)
{
	pad->priv->sticky = gtk_check_menu_item_get_active (check);
	if (pad->priv->sticky)
		gtk_window_stick (GTK_WINDOW (pad));
	else
		gtk_window_unstick (GTK_WINDOW (pad));
	xpad_pad_save_info_delayed (pad);
}

static void
menu_toolbar (XpadPad *pad, GtkCheckMenuItem *check)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) pad;

	g_object_set (xpad_global_settings, "has-toolbar", gtk_check_menu_item_get_active (check), NULL);
}

static void
menu_scrollbar (XpadPad *pad, GtkCheckMenuItem *check)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) pad;

	g_object_set (xpad_global_settings, "has-scrollbar", gtk_check_menu_item_get_active (check), NULL);
}

static void
menu_autohide (XpadPad *pad, GtkCheckMenuItem *check)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) pad;

	g_object_set (xpad_global_settings, "autohide-toolbar", gtk_check_menu_item_get_active (check), NULL);
}

static void
menu_decorated (XpadPad *pad, GtkCheckMenuItem *check)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) pad;

	g_object_set (xpad_global_settings, "has-decorations", gtk_check_menu_item_get_active (check), NULL);
}

static gint
menu_title_compare (GtkWindow *a, GtkWindow *b)
{
	gchar *title_a = g_utf8_casefold (gtk_window_get_title (a), -1);
	gchar *title_b = g_utf8_casefold (gtk_window_get_title (b), -1);
	
	gint rv = g_utf8_collate (title_a, title_b);
	
	g_free (title_a);
	g_free (title_b);
	
	return rv;
}

/* FIXME: Accelerators are working but not visible for menu items with an image (icon). */
#define MENU_ADD(mnemonic, image, key, mask, callback) {\
	if (image) {\
		item = gtk_menu_item_new ();\
		GtkWidget *hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 3);\
		GdkRGBA bg = {0, 0, 0, 0};\
		gtk_widget_override_background_color (hbox, GTK_STATE_FLAG_NORMAL, &bg);\
		gtk_container_add (GTK_CONTAINER (hbox), gtk_image_new_from_icon_name (image, GTK_ICON_SIZE_MENU));\
		gtk_container_add (GTK_CONTAINER (hbox), gtk_label_new_with_mnemonic (mnemonic));\
		gtk_container_add (GTK_CONTAINER (item), hbox);\
	}\
	else {\
		item = gtk_menu_item_new_with_mnemonic (mnemonic);\
	}\
	g_signal_connect_swapped (item, "activate", G_CALLBACK (callback), pad);\
	if (key)\
		gtk_widget_add_accelerator (item, "activate", accel_group, key, mask, GTK_ACCEL_VISIBLE);\
	gtk_container_add (GTK_CONTAINER (menu), item);\
	}

#define MENU_ADD_CHECK(mnemonic, active, callback) {\
	item = gtk_check_menu_item_new_with_mnemonic (mnemonic);\
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), active);\
	g_signal_connect_swapped (item, "toggled", G_CALLBACK (callback), pad);\
	gtk_container_add (GTK_CONTAINER (menu), item);\
	}

#define MENU_ADD_SEP() {\
	item = gtk_separator_menu_item_new ();\
	gtk_container_add (GTK_CONTAINER (menu), item);\
	}

static GtkWidget *
menu_get_popup_no_highlight (XpadPad *pad, GtkAccelGroup *accel_group)
{
	GtkWidget *uppermenu, *menu, *item;
	gboolean has_toolbar, autohide_toolbar, has_scrollbar, decorations;

	g_object_get (xpad_global_settings,
			"has-toolbar", &has_toolbar,
			"autohide-toolbar", &autohide_toolbar,
			"has-decorations", &decorations,
			"has-scrollbar", &has_scrollbar, NULL);
	
	uppermenu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (uppermenu), accel_group);

	/* Pad submenu */
	item = gtk_menu_item_new_with_mnemonic (_("_Pad"));
	gtk_container_add (GTK_CONTAINER (uppermenu), item);
	menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
	MENU_ADD (_("_New"), "document-new", 0, 0, xpad_pad_spawn);
	MENU_ADD_SEP ();
	MENU_ADD_CHECK (_("Show on _All Workspaces"), pad->priv->sticky, menu_sticky);
	g_object_set_data (G_OBJECT (uppermenu), "sticky", item);
	MENU_ADD (_("_Properties"), "document-properties", 0, 0, xpad_pad_open_properties);
	MENU_ADD_SEP ();
	MENU_ADD (_("_Close"), "window-close", 0, 0, xpad_pad_close);
	MENU_ADD (_("_Delete"), "edit-delete", 0, 0, xpad_pad_delete);

	/* Edit submenu */
	item = gtk_menu_item_new_with_mnemonic (_("_Edit"));
	gtk_container_add (GTK_CONTAINER (uppermenu), item);
	menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
	MENU_ADD (_("_Undo"), "edit-undo", GDK_KEY_Z, GDK_CONTROL_MASK, menu_undo);
	g_object_set_data (G_OBJECT (uppermenu), "undo", item);
	MENU_ADD (_("_Redo"), "edit-redo", GDK_KEY_Y, GDK_CONTROL_MASK, menu_redo);
	g_object_set_data (G_OBJECT (uppermenu), "redo", item);
	MENU_ADD_SEP();
	MENU_ADD (_("_Paste"), "edit-paste", 0, 0, menu_paste);
	g_object_set_data (G_OBJECT (uppermenu), "paste", item);
	MENU_ADD_SEP ();
	MENU_ADD (_("_Preferences"), "preferences-system", 0, 0, xpad_pad_open_preferences);

	/* View submenu */
	item = gtk_menu_item_new_with_mnemonic (_("_View"));
	gtk_container_add (GTK_CONTAINER (uppermenu), item);
	menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
	MENU_ADD_CHECK (_("_Toolbar"), has_toolbar, menu_toolbar);
	MENU_ADD_CHECK (_("_Autohide Toolbar"), autohide_toolbar, menu_autohide);
	gtk_widget_set_sensitive (item, has_toolbar);
	MENU_ADD_CHECK (_("_Scrollbar"), has_scrollbar, menu_scrollbar);
	MENU_ADD_CHECK (_("_Window Decorations"), decorations, menu_decorated);
	
	/* Notes submenu */
	item = gtk_menu_item_new_with_mnemonic (_("_Notes"));
	gtk_container_add (GTK_CONTAINER (uppermenu), item);
	menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
	g_object_set_data (G_OBJECT (uppermenu), "notes-menu", menu);
	MENU_ADD (_("_Show All"), NULL, 0, 0, menu_show_all);
	MENU_ADD (_("_Close All"), NULL, 0, 0, xpad_pad_close_all);

	/* Help submenu */
	item = gtk_menu_item_new_with_mnemonic (_("_Help"));
	gtk_container_add (GTK_CONTAINER (uppermenu), item);
	menu = gtk_menu_new ();
	gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), menu);
	MENU_ADD (_("_Help"), "help-browser", GDK_KEY_F1, 0, show_help);	
	MENU_ADD (_("_About"), "help-about", 0, 0, menu_about);
	
	gtk_widget_show_all (uppermenu);
	
	return uppermenu;
}

static void
menu_prep_popup_no_highlight (XpadPad *current_pad, GtkWidget *uppermenu)
{
	GtkWidget *menu, *item;

	GtkClipboard *clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);

	XpadTextBuffer *buffer = XPAD_TEXT_BUFFER (gtk_text_view_get_buffer (GTK_TEXT_VIEW (current_pad->priv->textview)));
	
	item = g_object_get_data (G_OBJECT (uppermenu), "paste");
	if (item)
		gtk_widget_set_sensitive (item, gtk_clipboard_wait_is_text_available (clipboard));

	item = g_object_get_data (G_OBJECT (uppermenu), "undo");
	if (item)
		gtk_widget_set_sensitive (item, xpad_text_buffer_undo_available (buffer));

	item = g_object_get_data (G_OBJECT (uppermenu), "redo");
	if (item)
		gtk_widget_set_sensitive (item, xpad_text_buffer_redo_available (buffer));
	
	menu = g_object_get_data (G_OBJECT (uppermenu), "notes-menu");
	if (menu)
	{
		gint n = 1;
		gchar *key;
		
		/* Remove old menu */
		item = g_object_get_data (G_OBJECT (menu), "notes-sep");
		while (item)
		{
			gtk_container_remove (GTK_CONTAINER (menu), item);
			key = g_strdup_printf ("notes-%i", n++);
			item = g_object_get_data (G_OBJECT (menu), key);
			g_free (key);
		}
	}
	if (menu && current_pad->priv->group)
	{
		GSList *pads, *l;
		gint n;
		GtkAccelGroup *accel_group = gtk_menu_get_accel_group (GTK_MENU (uppermenu));
		
		MENU_ADD_SEP ();
		g_object_set_data (G_OBJECT (menu), "notes-sep", item);
		
		/* Order pads according to title */
		pads = xpad_pad_group_get_pads (current_pad->priv->group);
		
		pads = g_slist_sort (pads, (GCompareFunc) menu_title_compare);
		
		/* Populate list of windows */
		for (l = pads, n = 1; l; l = l->next, n++)
		{
			gchar *title;
			gchar *tmp_title;
			gchar *key;
			GtkWidget *pad = GTK_WIDGET (l->data);
			
			key = g_strdup_printf ("notes-%i", n);
			tmp_title = g_strdup (gtk_window_get_title (GTK_WINDOW (pad)));
			str_replace_tokens (&tmp_title, '_', "__");
			if (n < 10)
				title = g_strdup_printf ("_%i. %s", n, tmp_title);
			else
				title = g_strdup_printf ("%i. %s", n, tmp_title);
			g_free (tmp_title);
			
			MENU_ADD (title, NULL, 0, 0, menu_show);
			g_object_set_data (G_OBJECT (menu), key, item);
			
			g_free (title);
			g_free (key);
		}
		g_slist_free (pads);
		g_slist_free (l);
		pads = NULL;
		l = NULL;
	}
}

static GtkWidget *
menu_get_popup_highlight (XpadPad *pad, GtkAccelGroup *accel_group)
{
	GtkWidget *menu, *item;
	
	menu = gtk_menu_new ();
	gtk_menu_set_accel_group (GTK_MENU (menu), accel_group);
	
	MENU_ADD (_("Cu_t"), "edit-cut", 0, 0, menu_cut);
	MENU_ADD (_("_Copy"), "edit-copy", 0, 0, menu_copy);	
	MENU_ADD (_("_Paste"), "edit-paste", 0, 0, menu_paste);	
	g_object_set_data (G_OBJECT (menu), "paste", item);
	MENU_ADD_SEP ();
	MENU_ADD (_("_Bold"), "format-text-bold", GDK_KEY_b, GDK_CONTROL_MASK, menu_bold);	
	MENU_ADD (_("_Italic"), "format-text-italic", GDK_KEY_i, GDK_CONTROL_MASK, menu_italic);	
	MENU_ADD (_("_Underline"), "format-text-underline", GDK_KEY_u, GDK_CONTROL_MASK, menu_underline);
	MENU_ADD (_("_Strikethrough"), "format-text-strikethrough", 0, 0, menu_strikethrough);	
	
	gtk_widget_show_all (menu);
	
	return menu;
}

static void
menu_prep_popup_highlight (XpadPad *pad, GtkWidget *menu)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) pad;

	GtkWidget *item;
	GtkClipboard *clipboard;
	
	clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
	
	item = g_object_get_data (G_OBJECT (menu), "paste");
	if (item)
		gtk_widget_set_sensitive (item, gtk_clipboard_wait_is_text_available (clipboard));
}

static void
menu_popup (GtkWidget *menu, XpadPad *pad)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) menu;

	g_signal_handlers_block_matched (pad, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) xpad_pad_leave_notify_event, NULL);
	pad->priv->toolbar_timeout = 0;
}

static void
menu_popdown (GtkWidget *menu, XpadPad *pad)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) menu;

	cairo_rectangle_int_t rect;

	/* We must check if we disabled off of pad and start the timeout if so. */
	rect.x = 10;
	rect.y = 10;
	rect.width = 1;
	rect.height = 1;
	
	if (!pad->priv->toolbar_timeout && !gtk_widget_intersect (GTK_WIDGET (pad), &rect, NULL))
		pad->priv->toolbar_timeout = g_timeout_add (1000, (GSourceFunc) toolbar_timeout, pad);

	g_signal_handlers_unblock_matched (pad, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) xpad_pad_leave_notify_event, NULL);
}

static void
xpad_pad_popup_deactivate (GtkWidget *menu, XpadPad *pad)
{
	menu_popdown (GTK_WIDGET (menu), pad);
}

static void
xpad_pad_toolbar_popup (GtkWidget *toolbar, GtkMenu *menu, XpadPad *pad)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) toolbar;

	menu_popup (GTK_WIDGET (menu), pad);
}

static void
xpad_pad_toolbar_popdown (GtkWidget *toolbar, GtkMenu *menu, XpadPad *pad)
{
	/* A dirty way to silence the compiler for these unused variables. */
	(void) toolbar;

	menu_popdown (GTK_WIDGET (menu), pad);
}

static void
xpad_pad_popup (XpadPad *pad, GdkEventButton *event)
{
	GtkTextBuffer *buffer;
	GtkWidget *menu;
	
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (pad->priv->textview));
	
	if (gtk_text_buffer_get_selection_bounds (buffer, NULL, NULL))
	{
		menu = pad->priv->highlight_menu;
		menu_prep_popup_highlight (pad, menu);
	}
	else
	{
		menu = pad->priv->menu;
		menu_prep_popup_no_highlight (pad, menu);
	}
	
	if (!menu)
		return;
	
	menu_popup (GTK_WIDGET (menu), pad);
	
	if (event)
		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, event->button, event->time);
	else
		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 0, gtk_get_current_event_time ());
}

/* These functions below are used to reduce the amounts of writes, hence improve the performance. */
void xpad_pad_save_content_delayed (XpadPad *pad)
{
   pad->priv->unsaved_content = TRUE;
   Xpad_periodic_save_content_delayed(pad);
}
void xpad_pad_save_info_delayed (XpadPad *pad)
{
   pad->priv->unsaved_info = TRUE;
   Xpad_periodic_save_info_delayed(pad);
}
void xpad_pad_save_unsaved (XpadPad *pad)
{
   if (pad->priv->unsaved_content)
      xpad_pad_save_content(pad);
   if (pad->priv->unsaved_info)
	   xpad_pad_save_info(pad);
}
